#include <gtk/gtk.h>
#include <gconf/gconf-client.h>
#include <string.h>
#include "blog-plugin.h"
#include "plugin-loader.h"
#include "accounts.h"

struct AccountsDialog {
	GtkWidget *url_label;
	GtkWidget *url_entry;
	GtkWidget *us_label;
	GtkWidget *us_entry;
	GtkWidget *pw_label;
	GtkWidget *pw_entry;
	GtkWidget *pl_label;
	GtkWidget *pl_combo;
	GtkWidget *tree_view;
	GtkWidget *remove;
	GtkListStore *list_store;
	GtkWidget *vbox;
	BlogPlugin **plugins;
};

enum Responses {
	RESPONSE_OK,
	RESPONSE_CANCEL,
	RESPONSE_ADD,
	RESPONSE_REMOVE
};

const gchar const *field_names[FIELD_COUNT] = {
	"url=",
	"user=",
	"password=",
	"plugin="
};

static gchar *_build_string(GtkTreeModel *model, GtkTreeIter *iter)
{
	GString *string = g_string_new("");
	enum Fields fld;

	for (fld = 0; fld < FIELD_COUNT; fld++) {
		gchar *txt;
		if (fld != FIELD_PLUGIN) {
			gtk_tree_model_get(model, iter, fld, &txt, -1);
		} else {
			BlogPlugin *plug;
			gtk_tree_model_get(model, iter, fld, &plug, -1);
			txt = plug ? g_strdup(plug->name) : NULL;
		}
		
		if (txt && *txt) {
			g_string_append(string, field_names[fld]);
			g_string_append(string, txt);
			g_string_append(string, "\n");
		}
		g_free(txt);
	}

	return g_strchomp(g_string_free(string, FALSE));
}

static void _cmb_changed(GtkWidget *entry,
		struct AccountsDialog *ad)
{
	GtkTreeIter iter;
	guint i;

	if (!gtk_tree_selection_get_selected(gtk_tree_view_get_selection(
					GTK_TREE_VIEW(ad->tree_view)),
				NULL,
				&iter)) {
		return;
	}

	i = gtk_combo_box_get_active(GTK_COMBO_BOX(entry));

	gtk_list_store_set(ad->list_store, &iter, FIELD_PLUGIN, ad->plugins[i], -1);

}

static void _changed(GtkWidget *entry,
		struct AccountsDialog *ad)
{
	gint col;
	const gchar *text;
	GtkTreeIter iter;

	if (!gtk_tree_selection_get_selected(gtk_tree_view_get_selection(
					GTK_TREE_VIEW(ad->tree_view)),
				NULL,
				&iter)) {
		return;
	}

	if (entry == ad->pw_entry) {
		col = FIELD_PASSWORD;
	} else if (entry == ad->us_entry) {
		col = FIELD_USER;
	} else {
		col = FIELD_URL;
	}

	text = gtk_entry_get_text(GTK_ENTRY(entry));
	gtk_list_store_set(ad->list_store, &iter, col, text, -1);
}

static void _account_changed(GtkTreeSelection *sel,
		struct AccountsDialog *ad)
{
	GtkTreeIter iter;
	GtkTreeModel *model;

	if (gtk_tree_selection_get_selected(sel, &model, &iter)) {
		gchar *url;
		gchar *user;
		gchar *password;
		BlogPlugin *plugin;

		gtk_tree_model_get(model, &iter,
			       	FIELD_URL, &url,
				FIELD_USER, &user,
				FIELD_PASSWORD, &password,
				FIELD_PLUGIN, &plugin,
				-1);

		if (plugin) {
			guint i;
			for (i = 0; ad->plugins[i] != NULL; i++) {
				if (plugin == ad->plugins[i]) {
					gtk_combo_box_set_active(GTK_COMBO_BOX(ad->pl_combo), i);
					break;
				}
			}
		} else {
			gtk_combo_box_set_active(GTK_COMBO_BOX(ad->pl_combo), 0);
			_cmb_changed(ad->pl_combo, ad);
		}

		gtk_entry_set_text(GTK_ENTRY(ad->url_entry), url ? url : "");
		gtk_entry_set_text(GTK_ENTRY(ad->us_entry), user ? user : "");
		gtk_entry_set_text(GTK_ENTRY(ad->pw_entry), password ? password : "");
		g_free(url);
		g_free(user);
		g_free(password);

		gtk_widget_set_sensitive(ad->vbox, TRUE);
		gtk_widget_set_sensitive(ad->remove, TRUE);
	} else {
		gtk_combo_box_set_active(GTK_COMBO_BOX(ad->pl_combo), 0);
		gtk_entry_set_text(GTK_ENTRY(ad->url_entry), "");
		gtk_entry_set_text(GTK_ENTRY(ad->us_entry), "");
		gtk_entry_set_text(GTK_ENTRY(ad->pw_entry), "");
		gtk_widget_set_sensitive(ad->vbox, FALSE);
		gtk_widget_set_sensitive(ad->remove, FALSE);
	}
}

GtkListStore *accounts_load(BlogPlugin **plugins)
{
	GConfClient *gc = gconf_client_get_default();
	GtkListStore *retval = gtk_list_store_new(FIELD_COUNT,
			G_TYPE_STRING,
			G_TYPE_STRING,
			G_TYPE_STRING,
			G_TYPE_POINTER);
	GSList *list = gconf_client_get_list(gc, "/apps/maemo/blog/accounts",
			GCONF_VALUE_STRING, NULL);
	GSList *liter;
	GtkTreeIter iter;

	for (liter = list; liter != NULL; liter = liter->next)
	{
		gchar **tokens;
		guint i;
		tokens = g_strsplit(liter->data, "\n", -1);

		gtk_list_store_append(retval,
				&iter);

		for (i = 0; tokens[i] != NULL; i++) {
			enum Fields fld;
			for (fld = 0; fld < FIELD_COUNT; fld++)
				if (g_str_has_prefix(tokens[i],
						       	field_names[fld]))
						break;
			switch (fld) {
			case FIELD_COUNT:
				continue;
				break;
			case FIELD_PLUGIN:
				gtk_list_store_set(retval,
						&iter,
						fld,
						plugins[plugin_id_by_name(plugins, tokens[i] + strlen(field_names[fld]))]);
				break;
			default:
				gtk_list_store_set(retval,
						&iter,
						fld,
						tokens[i] + strlen(field_names[fld]),
						-1);
				break;
			}
		}
	}

	g_slist_foreach(list, (GFunc)g_free, NULL);
	g_slist_free(list);

	return retval;
}

void account_name(GtkTreeViewColumn *tree_column,
		GtkCellRenderer *cell,
		GtkTreeModel *tree_model,
		GtkTreeIter *iter,
		gpointer data)
{
	gchar *user;
	const gchar *final_user;
	gchar *url;
	gchar *urlsep;
	const gchar *final_url;
	gchar *final_string;

	(void)tree_column;
	(void)data;

	gtk_tree_model_get(tree_model,
			iter,
			FIELD_USER, &user,
			FIELD_URL, &url,
			-1);

	if (user && *user) {
		final_user = user;
	} else {
		final_user = "(unknown)";
	}

	if (url && *url) {
		urlsep = strstr(url, "//");
		if (urlsep) {
			memmove(url, urlsep + 2, strlen(urlsep + 1));
		}
		urlsep = strchr(url, '/');
		if (urlsep) {
			*urlsep = '\0';
		}
		final_url = url;
	} else {
		final_url = "(unknown)";
	}

	final_string = g_strconcat(final_user, " @ ", final_url, NULL);
	g_object_set(cell, "text", final_string, NULL);

	g_free(final_string);
	g_free(user);
	g_free(url);
}

gboolean accounts_dialog(GtkWindow *parent, GtkListStore *accounts, BlogPlugin **plugins)
{
	GtkWidget *dialog;
	GtkWidget *hbox;
	GtkCellRenderer *renderer;
	GtkTreeIter iter;
	struct AccountsDialog ad;
	guint i;

	ad.plugins = plugins;
	ad.list_store = accounts;
	dialog = gtk_dialog_new_with_buttons("Accounts",
			parent,
			GTK_DIALOG_NO_SEPARATOR,
			NULL);
	gtk_dialog_add_button(GTK_DIALOG(dialog),
			"Ok",
			RESPONSE_OK);
	gtk_dialog_add_button(GTK_DIALOG(dialog),
			"Cancel",
			RESPONSE_CANCEL);
	gtk_dialog_add_button(GTK_DIALOG(dialog),
			"Add",
			RESPONSE_ADD);
	ad.remove = gtk_dialog_add_button(GTK_DIALOG(dialog),
			"Remove",
			RESPONSE_REMOVE);
	gtk_widget_set_sensitive(ad.remove, FALSE);

	hbox = gtk_hbox_new(FALSE, 0);
	ad.vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_set_sensitive(ad.vbox, FALSE);

	ad.tree_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(ad.list_store));
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(ad.tree_view), FALSE);

	g_signal_connect(gtk_tree_view_get_selection(GTK_TREE_VIEW(ad.tree_view)),
			"changed",
			G_CALLBACK(_account_changed),
			&ad);


	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(ad.tree_view),
			-1,
			"Name",
			renderer,
			account_name,
			NULL,
			NULL);

	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);
	gtk_container_add(GTK_CONTAINER(hbox), ad.tree_view);
	gtk_container_add(GTK_CONTAINER(hbox), ad.vbox);

	ad.url_label = gtk_label_new("Blog RPC URL");
	ad.url_entry = gtk_entry_new();
	g_signal_connect(ad.url_entry,
			"changed",
			G_CALLBACK(_changed),
			&ad);

	ad.us_label = gtk_label_new("Blog user");
	ad.us_entry = gtk_entry_new();
	g_signal_connect(ad.us_entry,
			"changed",
			G_CALLBACK(_changed),
			&ad);

	ad.pw_label = gtk_label_new("Blog password");
	ad.pw_entry = gtk_entry_new();
	gtk_entry_set_visibility(GTK_ENTRY(ad.pw_entry), FALSE);
	g_signal_connect(ad.pw_entry,
			"changed",
			G_CALLBACK(_changed),
			&ad);

	ad.pl_label = gtk_label_new("API");
	ad.pl_combo = gtk_combo_box_new_text();

	for (i = 0; plugins[i] != NULL; i++) {
		gtk_combo_box_append_text(GTK_COMBO_BOX(ad.pl_combo),
				plugins[i]->name);
	}
	g_signal_connect(ad.pl_combo,
			"changed",
			G_CALLBACK(_cmb_changed),
			&ad);


	gtk_container_add(GTK_CONTAINER(ad.vbox), ad.url_label);
	gtk_container_add(GTK_CONTAINER(ad.vbox), ad.url_entry);
	gtk_container_add(GTK_CONTAINER(ad.vbox), ad.us_label);
	gtk_container_add(GTK_CONTAINER(ad.vbox), ad.us_entry);
	gtk_container_add(GTK_CONTAINER(ad.vbox), ad.pw_label);
	gtk_container_add(GTK_CONTAINER(ad.vbox), ad.pw_entry);
	gtk_container_add(GTK_CONTAINER(ad.vbox), ad.pl_label);
	gtk_container_add(GTK_CONTAINER(ad.vbox), ad.pl_combo);

	gtk_widget_show_all(dialog);

	while (TRUE) {
		gint ret = gtk_dialog_run(GTK_DIALOG(dialog));
		if (ret == RESPONSE_OK) {
			gtk_widget_destroy(dialog);
			return TRUE;
		} else if (ret == RESPONSE_CANCEL) {
			gtk_widget_destroy(dialog);
			return FALSE;
		} else if (ret == RESPONSE_ADD) {
			gtk_list_store_append(ad.list_store,
					&iter);
			gtk_tree_selection_select_iter(
					gtk_tree_view_get_selection(
						GTK_TREE_VIEW(ad.tree_view)),
					&iter);
		} else if (ret == RESPONSE_REMOVE) {
			GtkTreeIter iter;
			GtkTreeModel *model;

			if (gtk_tree_selection_get_selected(
						gtk_tree_view_get_selection(
							GTK_TREE_VIEW(ad.tree_view)), &model, &iter)) {
				gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
			}
		}
	}

	return FALSE;
}

void accounts_save(GtkListStore *accounts)
{
	GSList *list = NULL;
	GtkTreeIter iter;
	GConfClient *gc = gconf_client_get_default();
	if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(accounts), &iter)) {
		gchar *temp = _build_string(GTK_TREE_MODEL(accounts), &iter);
		list = g_slist_append(list, temp);
		while (gtk_tree_model_iter_next(GTK_TREE_MODEL(accounts), &iter)) {
			temp = _build_string(GTK_TREE_MODEL(accounts), &iter);
			list = g_slist_append(list, temp);
		}
	}
	gconf_client_set_list(gc, "/apps/maemo/blog/accounts",
			GCONF_VALUE_STRING, list, NULL);
	g_slist_foreach(list, (GFunc)g_free, NULL);
	g_slist_free(list);
}

gboolean model_has_data(GtkTreeModel *model)
{
	GtkTreeIter iter;
	return gtk_tree_model_get_iter_first(model, &iter);
}
