#include <stdlib.h>
#include <stdio.h>

#include <glib.h>
#include <glib/gthread.h>
#include <libsoup/soup-session-async.h>
#include <libsoup/soup-xmlrpc-message.h>
#include <libsoup/soup-xmlrpc-response.h>
#include <string.h>

#include "blog-plugin.h"

struct BloggerPluginData {
	SoupSession *session;
};

struct BlogBloggerConnectedClosure {
	BlogPlugin *self;
	BlogConnectedCb cb;
	gpointer user_data;
	guint timeout_id;
	gchar *uri;
	gchar *user;
	gchar *pass;
};
static void _blog_blogger_connected(SoupMessage *message,
		gpointer user_data);
static gboolean blog_blogger_connect(BlogPlugin *plugin,
	       	const gchar *uri, const gchar *user,
		const gchar *pass, BlogConnectedCb cb, gpointer user_data);

struct BlogBloggerPostedClosure {
	Blog *blog;
	BlogPost *post;
	BlogPostCb cb;
	guint timeout_id;
	gpointer user_data;
	gboolean new_post;
};
static void _blog_blogger_posted(SoupMessage *message,
		gpointer user_data);
static gboolean blog_blogger_new_post(Blog *blog, BlogPost *post,
	       	gboolean publish, BlogPostCb cb, gpointer user_data);
static gboolean blog_blogger_edit_post(Blog *blog, BlogPost *post,
	       	gboolean publish, BlogPostCb cb, gpointer user_data);

struct BlogBloggerFetchedClosure {
	Blog *blog;
	BlogFetchCb cb;
	guint timeout_id;
	gpointer user_data;
};
static void blog_blogger_get_categories(Blog *blog, BlogFetchCb cb, gpointer user_data);
static void _blog_blogger_got_recent(SoupMessage *message,
		gpointer user_data);
static void blog_blogger_get_recent(Blog *blog, BlogFetchCb cb, gpointer user_data);
static void blog_blogger_unload(BlogPlugin *plugin);
static gboolean blog_blogger_upload_image(Blog *blog, const gchar *image_name, 
		const gchar *mime_type,
		const gchar *image_data,
		size_t data_size,
		BlogUploadCb cb,
		gpointer user_data);

BlogPlugin *plugin_init(void)
{
	BlogPlugin *retval = g_new0(BlogPlugin, 1);
	struct BloggerPluginData *data;

	retval->connect = blog_blogger_connect;
	retval->newPost = blog_blogger_new_post;
	retval->editPost = blog_blogger_edit_post;
	retval->getCategories = blog_blogger_get_categories;
	retval->getRecent = blog_blogger_get_recent;
	retval->unLoad = blog_blogger_unload;
	retval->uploadImage = blog_blogger_upload_image;

	retval->plugin_data = data = g_new0(struct BloggerPluginData, 1);
	retval->name = g_strdup("Blogger 1.0");

	if (!g_thread_supported()) {
		g_thread_init(NULL);
	}

	data->session = soup_session_async_new();

	return retval;
}

static void _blog_blogger_connected(SoupMessage *message,
		gpointer user_data)
{
	struct BlogBloggerConnectedClosure *clo = user_data;
	Blog **retval = NULL;
	SoupXmlrpcResponse *response = soup_xmlrpc_message_parse_response(
			SOUP_XMLRPC_MESSAGE(message));
	SoupXmlrpcValue *value = soup_xmlrpc_response_get_value(response);
	GSList *blogs = NULL;
	guint count = 0;
	SoupXmlrpcValueArrayIterator *iter;

	if (soup_xmlrpc_value_array_get_iterator(value, &iter)) {
		while (iter) {
			Blog *blog = blog_new();
			GHashTable *hash;
			SoupXmlrpcValue *val;
			SoupXmlrpcValue *array_val;

			soup_xmlrpc_value_array_iterator_get_value(iter,
				       	&array_val);

			soup_xmlrpc_value_get_struct(array_val, &hash);

			val = g_hash_table_lookup(hash, "blogid");
			soup_xmlrpc_value_get_string(val, &blog->id);
			val = g_hash_table_lookup(hash, "blogName");
			soup_xmlrpc_value_get_string(val, &blog->name);
			g_hash_table_destroy(hash);
			blog->url = g_strdup(clo->uri);
			blog->user = g_strdup(clo->user);
			blog->pass = g_strdup(clo->pass);
			blog->plugin = clo->self;

			blogs = g_slist_prepend(blogs, blog);
			count++;

			iter = soup_xmlrpc_value_array_iterator_next(iter);
		}
	} else {
			Blog *blog = blog_new();
			GHashTable *hash;
			SoupXmlrpcValue *val;
			soup_xmlrpc_value_get_struct(value, &hash);

			val = g_hash_table_lookup(hash, "blogid");
			soup_xmlrpc_value_get_string(val, &blog->id);
			val = g_hash_table_lookup(hash, "blogName");
			soup_xmlrpc_value_get_string(val, &blog->name);
			g_hash_table_destroy(hash);
			blog->url = g_strdup(clo->uri);
			blog->user = g_strdup(clo->user);
			blog->pass = g_strdup(clo->pass);
			blog->plugin = clo->self;

			blogs = g_slist_prepend(blogs, blog);
			count++;
	}

	retval = g_new0(Blog *, count + 1);
	retval[count] = NULL;

	while (blogs) {
		retval[--count] = blogs->data;
		blogs = g_slist_delete_link(blogs, blogs);
	}

	clo->cb(retval, NULL, clo->user_data);
	g_free(clo->uri);
	g_free(clo->user);
	g_free(clo->pass);
	g_free(clo);
}

static gboolean blog_blogger_connect(BlogPlugin *plugin,
		const gchar *uri, const gchar *user,
		const gchar *pass, BlogConnectedCb cb, gpointer user_data)
{
	struct BlogBloggerConnectedClosure *clo = g_new0(struct BlogBloggerConnectedClosure, 1);
	struct BloggerPluginData *data = plugin->plugin_data;
	SoupXmlrpcMessage *msg;
	clo->cb = cb;
	clo->user_data = user_data;
	clo->uri = g_strdup(uri);
	clo->user = g_strdup(user);
	clo->pass = g_strdup(pass);
	clo->self = plugin;

	msg = soup_xmlrpc_message_new(uri);
	soup_xmlrpc_message_start_call(msg,
			"blogger.getUsersBlogs");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg,
			"0123456789ABCDEF");
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg,
			user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg,
			pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg),
			_blog_blogger_connected, clo);

	return TRUE;
}
static void _blog_blogger_posted(SoupMessage *message,
		gpointer user_data)
{
	struct BlogBloggerPostedClosure *clo = user_data;
	gboolean status = TRUE;
	long numeric_id = 0;
	gchar *string_id = NULL;

	SoupXmlrpcResponse *response = soup_xmlrpc_message_parse_response(
			SOUP_XMLRPC_MESSAGE(message));

	SoupXmlrpcValue *value = soup_xmlrpc_response_get_value(response);


	if (clo->new_post) {
		switch (soup_xmlrpc_value_get_type(value)) {
		case SOUP_XMLRPC_VALUE_TYPE_INT:
			soup_xmlrpc_value_get_int(value,
					&numeric_id);
			break;
		case SOUP_XMLRPC_VALUE_TYPE_STRING:
			soup_xmlrpc_value_get_string(value,
					&string_id);
			break;
		default:
			/* ??? */
			break;
		}

		if (string_id) {
			clo->post->id = string_id;
		} else {
			clo->post->id = g_strdup_printf("%ld\n", numeric_id);
		}

		clo->blog->posts = g_slist_prepend(clo->blog->posts, clo->post);
	} else {
		soup_xmlrpc_value_get_boolean(value, &status);
	}
	clo->cb(clo->blog, clo->post, status, NULL, clo->user_data);
	g_free(clo);
}


static gboolean blog_blogger_new_post(Blog *blog, BlogPost *post,
	       	gboolean publish, BlogPostCb cb, gpointer user_data)
{
	struct BlogBloggerPostedClosure *clo;
	struct BloggerPluginData *data = blog->plugin->plugin_data;
	SoupXmlrpcMessage *msg;
	gchar *content;

	if (!post || !post->content) {
		return FALSE;
	}

	clo = g_new0(struct BlogBloggerPostedClosure, 1);
	clo->blog = blog;
	clo->post = post;
	clo->new_post = TRUE;
	clo->cb = cb;
	clo->user_data = user_data;

	content = g_strdup_printf("<title>%s</title>\n%s",
			post->title ? post->title : "",
			post->content);

	msg = soup_xmlrpc_message_new(blog->url);
	soup_xmlrpc_message_start_call(msg, "blogger.newPost");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, "0123456789ABCDEF");
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->id);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, content);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_boolean(msg, publish);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg),
			_blog_blogger_posted, clo);

	g_free(content);

	return TRUE;
}

static gboolean blog_blogger_edit_post(Blog *blog, BlogPost *post,
		gboolean publish, BlogPostCb cb, gpointer user_data)
{
	gchar *content;
	struct BlogBloggerPostedClosure *clo;
	struct BloggerPluginData *data = blog->plugin->plugin_data;
	SoupXmlrpcMessage *msg;

	if (!post || !post->content || !post->id) {
		return FALSE;
	}

	clo = g_new0(struct BlogBloggerPostedClosure, 1);
	clo->blog = blog;
	clo->post = post;
	clo->new_post = FALSE;
	clo->cb = cb;
	clo->user_data = user_data;

	content = g_strdup_printf("<title>%s</title>\n%s",
			post->title ? post->title : "",
			post->content);

	msg = soup_xmlrpc_message_new(blog->url);
	soup_xmlrpc_message_start_call(msg, "blogger.editPost");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, "0123456789ABCDEF");
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, post->id);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, content);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_boolean(msg, publish);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg),
			_blog_blogger_posted, clo);

	g_free(content);

	return TRUE;
}

static gboolean _blog_blogger_got_categories(gpointer user_data)
{
	struct BlogBloggerFetchedClosure *clo = user_data;

	clo->cb(clo->blog, NULL, clo->user_data);

	return FALSE;
}

static void blog_blogger_get_categories(Blog *blog, BlogFetchCb cb, gpointer user_data)
{
	struct BlogBloggerFetchedClosure *clo = g_new0(struct BlogBloggerFetchedClosure, 1);

	blog->categories = NULL;

	clo->blog = blog;
	clo->cb = cb;
	clo->user_data = user_data;

	g_idle_add(_blog_blogger_got_categories, clo);
}

static void _blog_blogger_got_recent(SoupMessage *message,
		gpointer user_data)
{
	struct BlogBloggerFetchedClosure *clo = (struct BlogBloggerFetchedClosure *)user_data;
	SoupXmlrpcResponse *response = soup_xmlrpc_message_parse_response(
			SOUP_XMLRPC_MESSAGE(message));
	SoupXmlrpcValue *array = soup_xmlrpc_response_get_value(response);
	SoupXmlrpcValue *value;
	SoupXmlrpcValueArrayIterator *iter;

	for (soup_xmlrpc_value_array_get_iterator(array, &iter);
			iter;
			iter = soup_xmlrpc_value_array_iterator_next(iter)) {
		BlogPost *new_post = g_new0(BlogPost, 1);
		soup_xmlrpc_value_array_iterator_get_value(iter, &value);
		gchar *id, *content;
		gchar *temp1, *temp2;
		GHashTable *table;

		soup_xmlrpc_value_get_struct(value, &table);
		value = g_hash_table_lookup(table, "postid");
		if (value) {
			soup_xmlrpc_value_get_string(value, &id);
		}
		value = g_hash_table_lookup(table, "content");
		if (value) {
			soup_xmlrpc_value_get_string(value, &content);
		}

		new_post->id = id;
		temp1 = strstr(content, "<title>");
		temp2 = strstr(content, "</title>");

		if (temp1 && temp2) {
			temp1 += strlen("<title>");
			new_post->title = g_strndup(temp1, temp2 - temp1);
		}
		if (temp2) {
			temp2 += strlen("</title>");
			new_post->content = g_strdup(temp2);
			g_free(content);
		} else {
			new_post->content = content;
		}

		g_hash_table_destroy(table);

		clo->blog->posts = g_slist_append(clo->blog->posts, new_post);
	}

	clo->cb(clo->blog, NULL, clo->user_data);
	g_free(clo);
}

static void blog_blogger_get_recent(Blog *blog, BlogFetchCb cb, gpointer user_data)
{
	struct BlogBloggerFetchedClosure *clo = g_new0(struct BlogBloggerFetchedClosure, 1);
	struct BloggerPluginData *data = blog->plugin->plugin_data;
	SoupXmlrpcMessage *msg;
	clo->blog = blog;
	clo->cb = cb;
	clo->user_data = user_data;

	msg = soup_xmlrpc_message_new(blog->url);
	soup_xmlrpc_message_start_call(msg, "metaWeblog.getRecentPosts");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, "0123456789ABCDEF");
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->id);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_int(msg, 15);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg),
			_blog_blogger_got_recent, clo);
}

static void blog_blogger_unload(BlogPlugin *plugin)
{
	struct BloggerPluginData *data = plugin->plugin_data;
	g_object_unref(data->session);
	g_free(data);
	g_free(plugin->name);
	g_free(plugin);
}

static gboolean blog_blogger_upload_image(Blog *blog, const gchar *image_name, 
		const gchar *mime_type,
		const gchar *image_data,
		size_t data_size, BlogUploadCb cb, gpointer user_data)
{
	(void)blog;
	(void)image_name;
	(void)image_data;
	(void)mime_type;
	(void)data_size;
	(void)cb;
	(void)user_data;

	return FALSE;
}
