#include <stdlib.h>
#include <stdio.h>
#include <glib.h>
#include <glib/gthread.h>
#include <libsoup/soup-session-async.h>
#include <libsoup/soup-xmlrpc-message.h>
#include <libsoup/soup-xmlrpc-response.h>
#include <string.h>

#include "blog-plugin.h"

struct MetaweblogPluginData {
	SoupSession *session;
};

static gint _field_compare(gconstpointer a, gconstpointer b)
{
	const BlogParameter *pa = a;
	const BlogParameter *pb = b;

	return strcmp(pa->name, pb->name);
}

static BlogCategory *category_lookup(Blog *blog, const gchar *name)
{
	GSList *iter;

	for (iter = blog->categories; iter; iter = iter->next)
	{
		BlogCategory *cat = (BlogCategory *)iter->data;

		if (strcmp(cat->name, name) == 0) {
			return iter->data;
		}
	}

	return NULL;
}

struct BlogMetaweblogConnectedClosure {
	BlogPlugin *self;
	BlogConnectedCb cb;
	gpointer user_data;
	guint timeout_id;
	gchar *uri;
	gchar *user;
	gchar *pass;
};
static void _blog_metaweblog_connected(SoupMessage *message,
		gpointer user_data);
static gboolean blog_metaweblog_connect(BlogPlugin *plugin,
	       	const gchar *uri, const gchar *user,
		const gchar *pass, BlogConnectedCb cb, gpointer user_data);

struct BlogMetaweblogPostedClosure {
	Blog *blog;
	BlogPost *post;
	BlogPostCb cb;
	guint timeout_id;
	gpointer user_data;
	gboolean new_post;
};
static void _blog_metaweblog_posted(SoupMessage *message,
		gpointer user_data);
static gboolean blog_metaweblog_new_post(Blog *blog, BlogPost *post,
	       	gboolean publish, BlogPostCb cb, gpointer user_data);
static gboolean blog_metaweblog_edit_post(Blog *blog, BlogPost *post,
	       	gboolean publish, BlogPostCb cb, gpointer user_data);

struct BlogMetaweblogFetchedClosure {
	Blog *blog;
	BlogFetchCb cb;
	guint timeout_id;
	gpointer user_data;
};

static void _blog_metaweblog_got_categories(SoupMessage *message,
		gpointer user_data);
static void blog_metaweblog_get_categories(Blog *blog, BlogFetchCb cb, gpointer user_data);
static void _blog_metaweblog_got_recent(SoupMessage *message,
		gpointer user_data);
static void blog_metaweblog_get_recent(Blog *blog, BlogFetchCb cb, gpointer user_data);
static void blog_metaweblog_unload(BlogPlugin *plugin);

struct BlogMetaweblogUploadedClosure {
	Blog *blog;
	BlogUploadCb cb;
	guint timeout_id;
	gchar *image_name;
	gpointer user_data;
	gboolean new_post;
};
static void _blog_metaweblog_uploaded(SoupMessage *message,
		gpointer user_data);
static gboolean blog_metaweblog_upload_image(Blog *blog, const gchar *image_name, 
		const gchar *mime_type,
		const gchar *image_data,
		size_t data_size,
		BlogUploadCb cb,
		gpointer user_data);

BlogPlugin *plugin_init(void)
{
	BlogPlugin *retval = g_new0(BlogPlugin, 1);
	struct MetaweblogPluginData *data;

	retval->connect = blog_metaweblog_connect;
	retval->newPost = blog_metaweblog_new_post;
	retval->editPost = blog_metaweblog_edit_post;
	retval->getCategories = blog_metaweblog_get_categories;
	retval->getRecent = blog_metaweblog_get_recent;
	retval->unLoad = blog_metaweblog_unload;
	retval->uploadImage = blog_metaweblog_upload_image;

	retval->name = g_strdup("MetaWebLog");
	retval->plugin_data = data = g_new0(struct MetaweblogPluginData, 1);

	if (!g_thread_supported()) {
		g_thread_init(NULL);
	}

	data->session = soup_session_async_new();
	return retval;
}

static void _blog_metaweblog_connected(SoupMessage *message,
		gpointer user_data)
{
	struct BlogMetaweblogConnectedClosure *clo = (struct BlogMetaweblogConnectedClosure *)user_data;
	Blog **retval = NULL;
	SoupXmlrpcResponse *response = soup_xmlrpc_message_parse_response(
			SOUP_XMLRPC_MESSAGE(message));

	SoupXmlrpcValue *value = soup_xmlrpc_response_get_value(response);
	GSList *blogs = NULL;
	guint count = 0;

	SoupXmlrpcValueArrayIterator *iter;

	if (soup_xmlrpc_response_is_fault(response)) {
		BlogPluginError error = { 0, NULL };
		error.error_string = "Unknown error";
		error.error_code = 404;
		clo->cb(retval, &error, clo->user_data);
		g_free(clo->uri);
		g_free(clo->user);
		g_free(clo->pass);
		g_free(clo);
		return;
	}

	soup_xmlrpc_value_array_get_iterator(value, &iter);

	while (iter) {
		Blog *blog = blog_new();
		GHashTable *hash;
		SoupXmlrpcValue *val;
		SoupXmlrpcValue *array_val;

		soup_xmlrpc_value_array_iterator_get_value(iter, &array_val);

		soup_xmlrpc_value_get_struct(array_val, &hash);

		val = g_hash_table_lookup(hash, "blogid");
		soup_xmlrpc_value_get_string(val, &blog->id);
		val = g_hash_table_lookup(hash, "blogName");
		soup_xmlrpc_value_get_string(val, &blog->name);
		blog->url = g_strdup(clo->uri);
		blog->user = g_strdup(clo->user);
		blog->pass = g_strdup(clo->pass);
		blog->plugin = clo->self;

		blogs = g_slist_prepend(blogs, blog);
		count++;

		iter = soup_xmlrpc_value_array_iterator_next(iter);
	}

	retval = g_new0(Blog *, count + 1);
	retval[count] = NULL;

	while (blogs) {
		retval[--count] = blogs->data;
		blogs = g_slist_delete_link(blogs, blogs);
	}

	clo->cb(retval, NULL, clo->user_data);
	g_free(clo->uri);
	g_free(clo->user);
	g_free(clo->pass);
	g_free(clo);
}

static gboolean blog_metaweblog_connect(BlogPlugin *plugin,
		const gchar *uri, const gchar *user,
		const gchar *pass, BlogConnectedCb cb, gpointer user_data)
{
	struct BlogMetaweblogConnectedClosure *clo = g_new0(struct BlogMetaweblogConnectedClosure, 1);
	struct MetaweblogPluginData *data = plugin->plugin_data;
	SoupXmlrpcMessage *msg;

	clo->cb = cb;
	clo->user_data = user_data;
	clo->uri = g_strdup(uri);
	clo->user = g_strdup(user);
	clo->pass = g_strdup(pass);
	clo->self = plugin;

	msg = soup_xmlrpc_message_new(uri);
	soup_xmlrpc_message_start_call(msg,
			"blogger.getUsersBlogs");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg,
			"0123456789ABCDEF");
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg,
			user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg,
			pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg),
			_blog_metaweblog_connected, clo);

	return TRUE;
}

static void _blog_metaweblog_posted(SoupMessage *message,
		gpointer user_data)
{
	struct BlogMetaweblogPostedClosure *clo = (struct BlogMetaweblogPostedClosure *)user_data;
	gboolean status = TRUE;
	long numeric_id = 0;
	gchar *string_id = NULL;
	SoupXmlrpcResponse *response = soup_xmlrpc_message_parse_response(
			SOUP_XMLRPC_MESSAGE(message));

	SoupXmlrpcValue *value = soup_xmlrpc_response_get_value(response);

	if (clo->new_post) {
		switch (soup_xmlrpc_value_get_type(value)) {
		case SOUP_XMLRPC_VALUE_TYPE_INT:
			soup_xmlrpc_value_get_int(value,
					&numeric_id);
			break;
		case SOUP_XMLRPC_VALUE_TYPE_STRING:
			soup_xmlrpc_value_get_string(value,
					&string_id);
			break;
		default:
			/* ??? */
			break;
		}

		if (string_id) {
			clo->post->id = string_id;
		} else {
			clo->post->id = g_strdup_printf("%ld\n", numeric_id);
		}

		clo->blog->posts = g_slist_prepend(clo->blog->posts, clo->post);
	} else {
		soup_xmlrpc_value_get_boolean(value, &status);
	}
	clo->cb(clo->blog, clo->post, status, NULL, clo->user_data);
	g_free(clo);
}

static gboolean blog_metaweblog_new_post(Blog *blog, BlogPost *post,
	       	gboolean publish, BlogPostCb cb, gpointer user_data)
{
	GSList *iter;
	struct BlogMetaweblogPostedClosure *clo;
	struct MetaweblogPluginData *data = blog->plugin->plugin_data;
	SoupXmlrpcMessage *msg;
	GSList *param_copy = NULL;
	GSList *struct_stack = g_slist_prepend(NULL, g_strdup(""));
	guint top_len = 0;

	if (!post || !post->content) {
		return FALSE;
	}

	clo = g_new0(struct BlogMetaweblogPostedClosure, 1);
	clo->blog = blog;
	clo->post = post;
	clo->new_post = TRUE;
	clo->cb = cb;
	clo->user_data = user_data;

	msg = soup_xmlrpc_message_new(blog->url);
	soup_xmlrpc_message_start_call(msg, "metaWeblog.newPost");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->id);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_start_struct(msg);
	soup_xmlrpc_message_start_member(msg,
			"title");
	soup_xmlrpc_message_write_string(msg,
			post->title);
	soup_xmlrpc_message_end_member(msg);
	soup_xmlrpc_message_start_member(msg,
			"description");
	soup_xmlrpc_message_write_string(msg,
			post->content);
	soup_xmlrpc_message_end_member(msg);
	soup_xmlrpc_message_start_member(msg,
			"category");
	soup_xmlrpc_message_start_array(msg);
	for (iter = post->categories; iter; iter = iter->next)
	{
		BlogCategory *cat = (BlogCategory *)iter->data;
		soup_xmlrpc_message_write_string(msg, cat->id);
	}
	soup_xmlrpc_message_end_array(msg);
	soup_xmlrpc_message_end_member(msg);
	soup_xmlrpc_message_start_member(msg,
			"categories");
	soup_xmlrpc_message_start_array(msg);
	for (iter = post->categories; iter; iter = iter->next)
	{
		BlogCategory *cat = (BlogCategory *)iter->data;
		soup_xmlrpc_message_write_string(msg, cat->id);
	}
	soup_xmlrpc_message_end_array(msg);
	soup_xmlrpc_message_end_member(msg);
	param_copy = g_slist_sort(g_slist_copy(post->extra_fields),
			_field_compare);
	for (iter = param_copy; iter; iter = iter->next) {
		BlogParameter *parm = iter->data;
		gchar *temp = parm->name;
		while (strncmp(struct_stack->data, parm->name,
					top_len)) {
			soup_xmlrpc_message_end_struct(msg);
			soup_xmlrpc_message_end_member(msg);
			g_free(struct_stack->data);
			struct_stack = g_slist_delete_link(struct_stack,
					struct_stack);
		}
		while ((temp = strchr(parm->name + top_len,
					       	':'))) {
			gchar *struct_name =
				g_strndup(parm->name +
						top_len,
					       	temp - parm->name -
					       	top_len);
			soup_xmlrpc_message_start_member(msg, struct_name);
			g_free(struct_name);
			top_len = temp - parm->name + 1;
			struct_name = g_strndup(parm->name, top_len);
			soup_xmlrpc_message_start_struct(msg);
			struct_stack = g_slist_prepend(struct_stack,
					struct_name);
		}
		soup_xmlrpc_message_start_member(msg, parm->name + strlen(struct_stack->data));
		switch (G_VALUE_TYPE(&parm->value)) {
			case G_TYPE_STRING:
				soup_xmlrpc_message_write_string(msg,
						g_value_get_string(
							&parm->value));
				break;
			case G_TYPE_INT:
				soup_xmlrpc_message_write_int(msg,
						g_value_get_int(
							&parm->value));
				break;
			case G_TYPE_UINT:
				soup_xmlrpc_message_write_int(msg,
						(int)g_value_get_uint(
							&parm->value));
				break;
			case G_TYPE_BOOLEAN:
				soup_xmlrpc_message_write_boolean(msg,
						g_value_get_boolean(
							&parm->value));
				break;
			default:
				break;
		}
		soup_xmlrpc_message_end_member(msg);
	}
	while (struct_stack) {
		soup_xmlrpc_message_end_struct(msg);
		g_free(struct_stack->data);
		struct_stack = g_slist_delete_link(struct_stack, struct_stack);
	}
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_boolean(msg, publish);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg),
			_blog_metaweblog_posted, clo);

	return TRUE;
}

static gboolean blog_metaweblog_edit_post(Blog *blog, BlogPost *post,
		gboolean publish, BlogPostCb cb, gpointer user_data)
{
	GSList *iter;
	struct BlogMetaweblogPostedClosure *clo;
	struct MetaweblogPluginData *data = blog->plugin->plugin_data;
	SoupXmlrpcMessage *msg;

	if (!post || !post->content || !post->id) {
		return FALSE;
	}

	clo = g_new0(struct BlogMetaweblogPostedClosure, 1);
	clo->blog = blog;
	clo->post = post;
	clo->cb = cb;
	clo->user_data = user_data;
	clo->new_post = FALSE;

	msg = soup_xmlrpc_message_new(blog->url);
	soup_xmlrpc_message_start_call(msg, "metaWeblog.editPost");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, post->id);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_start_struct(msg);
	soup_xmlrpc_message_start_member(msg,
			"title");
	soup_xmlrpc_message_write_string(msg,
			post->title);
	soup_xmlrpc_message_end_member(msg);
	soup_xmlrpc_message_start_member(msg,
			"description");
	soup_xmlrpc_message_write_string(msg,
			post->content);
	soup_xmlrpc_message_end_member(msg);
	soup_xmlrpc_message_start_member(msg,
			"category");
	soup_xmlrpc_message_start_array(msg);
	for (iter = post->categories; iter; iter = iter->next)
	{
		BlogCategory *cat = (BlogCategory *)iter->data;
		soup_xmlrpc_message_write_string(msg, cat->id);
	}
	soup_xmlrpc_message_end_array(msg);
	soup_xmlrpc_message_end_member(msg);
	soup_xmlrpc_message_start_member(msg,
			"categories");
	soup_xmlrpc_message_start_array(msg);
	for (iter = post->categories; iter; iter = iter->next)
	{
		BlogCategory *cat = (BlogCategory *)iter->data;
		soup_xmlrpc_message_write_string(msg, cat->id);
	}
	soup_xmlrpc_message_end_array(msg);
	soup_xmlrpc_message_end_member(msg);
	for (iter = post->extra_fields; iter; iter = iter->next) {
		BlogParameter *parm = iter->data;
		soup_xmlrpc_message_start_member(msg, parm->name);
		switch (G_VALUE_TYPE(&parm->value)) {
			case G_TYPE_STRING:
				soup_xmlrpc_message_write_string(msg,
						g_value_get_string(
							&parm->value));
				break;
			case G_TYPE_INT:
				soup_xmlrpc_message_write_int(msg,
						g_value_get_int(
							&parm->value));
				break;
			case G_TYPE_UINT:
				soup_xmlrpc_message_write_int(msg,
						(int)g_value_get_uint(
							&parm->value));
				break;
			case G_TYPE_BOOLEAN:
				soup_xmlrpc_message_write_boolean(msg,
						g_value_get_boolean(
							&parm->value));
				break;
			default:
				break;
		}
		soup_xmlrpc_message_end_member(msg);
	}
	soup_xmlrpc_message_end_struct(msg);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_boolean(msg, publish);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg),
			_blog_metaweblog_posted, clo);

	return TRUE;
}

static void _add_category(gpointer key, gpointer value, gpointer user_data)
{
	GSList **categories = user_data;
	gchar *id = key;
	gchar *name = NULL;
	SoupXmlrpcValue *val = value;
	GHashTable *table;
	BlogCategory *category =
		g_new0(BlogCategory, 1);

	soup_xmlrpc_value_get_struct(val, &table);
	val = g_hash_table_lookup(table, "description");
	if (val) {
		soup_xmlrpc_value_get_string(val, &name);
	}

	category->id = g_strdup(id);
	category->name = name ? name : g_strdup(id);

	*categories = g_slist_append(*categories, category);
}

static void _blog_metaweblog_got_categories(SoupMessage *message,
		gpointer user_data)
{
	struct BlogMetaweblogFetchedClosure *clo = (struct BlogMetaweblogFetchedClosure *)user_data;

	SoupXmlrpcResponse *response = soup_xmlrpc_message_parse_response(
			SOUP_XMLRPC_MESSAGE(message));
	SoupXmlrpcValue *value = soup_xmlrpc_response_get_value(response);

	if (soup_xmlrpc_value_get_type(value) == SOUP_XMLRPC_VALUE_TYPE_ARRAY) {
		SoupXmlrpcValueArrayIterator *iter;
		SoupXmlrpcValue *subvalue;

		soup_xmlrpc_value_array_get_iterator(value, &iter);

		while (iter) {
			soup_xmlrpc_value_array_iterator_get_value(iter,&subvalue);
			GHashTable *table;
			gchar *id = NULL, *name = NULL;
			soup_xmlrpc_value_get_struct(subvalue, &table);

			subvalue = g_hash_table_lookup(table,
					"categoryId");
			if (subvalue) {
				soup_xmlrpc_value_get_string(subvalue,
						&id);
			}
			subvalue = g_hash_table_lookup(table,
					"description");
			if (subvalue) {
				soup_xmlrpc_value_get_string(subvalue,
						&name);
			}
			if (name) {
				BlogCategory *category =
					g_new0(BlogCategory, 1);
				category->id = id ? id : g_strdup(name);
				category->name = name;

				clo->blog->categories = g_slist_append(
						clo->blog->categories,
						category);
			}
			iter = soup_xmlrpc_value_array_iterator_next(iter);
		}
	} else {
		GHashTable *table;
		soup_xmlrpc_value_get_struct(value, &table);
		g_hash_table_foreach(table, _add_category,
				&clo->blog->categories);
	}

	clo->cb(clo->blog, NULL, clo->user_data);
	g_free(clo);
}

static void blog_metaweblog_get_categories(Blog *blog, BlogFetchCb cb, gpointer user_data)
{
	struct BlogMetaweblogFetchedClosure *clo = g_new0(struct BlogMetaweblogFetchedClosure, 1);
	struct MetaweblogPluginData *data = blog->plugin->plugin_data;
	SoupXmlrpcMessage *msg;
	clo->blog = blog;
	clo->cb = cb;
	clo->user_data = user_data;

	msg = soup_xmlrpc_message_new(blog->url);
	soup_xmlrpc_message_start_call(msg, "metaWeblog.getCategories");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->id);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg),
			_blog_metaweblog_got_categories, clo);
}

static void _blog_metaweblog_got_recent(SoupMessage *message,
		gpointer user_data)
{
	struct BlogMetaweblogFetchedClosure *clo = (struct BlogMetaweblogFetchedClosure *)user_data;
	SoupXmlrpcResponse *response = soup_xmlrpc_message_parse_response(
			SOUP_XMLRPC_MESSAGE(message));
	SoupXmlrpcValue *array = soup_xmlrpc_response_get_value(response);
	SoupXmlrpcValue *value;
	SoupXmlrpcValueArrayIterator *iter;

	for (soup_xmlrpc_value_array_get_iterator(array, &iter);
			iter;
			iter = soup_xmlrpc_value_array_iterator_next(iter)) {
		soup_xmlrpc_value_array_iterator_get_value(iter, &value);
		BlogPost *new_post = g_new0(BlogPost, 1);
		GHashTable *table;
		gchar *id, *content;
		gchar *title;

		soup_xmlrpc_value_get_struct(value, &table);
		value = g_hash_table_lookup(table, "postid");
		if (value) {
			soup_xmlrpc_value_get_string(value, &id);
		}
		value = g_hash_table_lookup(table, "title");
		if (value) {
			soup_xmlrpc_value_get_string(value, &title);
		}
		value = g_hash_table_lookup(table, "description");
		if (value) {
			soup_xmlrpc_value_get_string(value, &content);
		}
		value = g_hash_table_lookup(table, "categories");
		if (!value) {
			value = g_hash_table_lookup(table, "category");
		}
		if (value) {
			SoupXmlrpcValueArrayIterator *sub_iter;
			soup_xmlrpc_value_dump(value);
			for (soup_xmlrpc_value_array_get_iterator(value, &sub_iter);
					sub_iter;
					sub_iter = 
					soup_xmlrpc_value_array_iterator_next(
						sub_iter)) {
				gchar *catid = NULL;
				BlogCategory *cat;
				if (!soup_xmlrpc_value_array_iterator_get_value(sub_iter, &value)) {
				}
				if (!soup_xmlrpc_value_get_string(value, &catid)) {
				}
				cat = category_lookup(clo->blog, catid);
				if (cat) {
					new_post->categories = g_slist_append(
							new_post->categories,
							cat);
				}
				g_free(catid);
			}
		}

		g_hash_table_destroy(table);

		new_post->id = id;
		new_post->content = content;
		new_post->title = title;

		clo->blog->posts = g_slist_append(clo->blog->posts, new_post);
	}

	clo->cb(clo->blog, NULL, clo->user_data);
	g_free(clo);
}

static void blog_metaweblog_get_recent(Blog *blog, BlogFetchCb cb, gpointer user_data)
{
	struct BlogMetaweblogFetchedClosure *clo = g_new0(struct BlogMetaweblogFetchedClosure, 1);
	struct MetaweblogPluginData *data = blog->plugin->plugin_data;
	SoupXmlrpcMessage *msg;
	clo->blog = blog;
	clo->cb = cb;
	clo->user_data = user_data;

	msg = soup_xmlrpc_message_new(blog->url);
	soup_xmlrpc_message_start_call(msg, "metaWeblog.getRecentPosts");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->id);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_int(msg, 15);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg),
			_blog_metaweblog_got_recent, clo);
}

static void blog_metaweblog_unload(BlogPlugin *plugin)
{
	struct MetaweblogPluginData *data = plugin->plugin_data;
	g_object_unref(data->session);
	g_free(data);
	g_free(plugin->name);
	g_free(plugin);
}

static void _blog_metaweblog_uploaded(SoupMessage *message,
		gpointer user_data)
{
	struct BlogMetaweblogUploadedClosure *clo = (struct BlogMetaweblogUploadedClosure *)user_data;
	SoupXmlrpcResponse *response = soup_xmlrpc_message_parse_response(
			SOUP_XMLRPC_MESSAGE(message));
	GHashTable *table;
	gchar *url = NULL;
	SoupXmlrpcValue *value = soup_xmlrpc_response_get_value(response);

	soup_xmlrpc_value_get_struct(value, &table);
	value = g_hash_table_lookup(table, "url");

	soup_xmlrpc_value_get_string(value, &url);

	clo->cb(clo->blog, clo->image_name, url, NULL, clo->user_data);
	g_free(url);
	g_free(clo->image_name);
	g_free(clo);
}

static gboolean blog_metaweblog_upload_image(Blog *blog, const gchar *image_name, 
			const gchar *mime_type, const gchar *image_data,
			size_t data_size, BlogUploadCb cb, gpointer user_data)
{
	struct BlogMetaweblogUploadedClosure *clo = g_new0(struct BlogMetaweblogUploadedClosure, 1);
	struct MetaweblogPluginData *data = blog->plugin->plugin_data;
	SoupXmlrpcMessage *msg;

	clo->cb = cb;
	clo->user_data = user_data;
	clo->image_name = g_strdup(image_name);
	clo->blog = blog;

	msg = soup_xmlrpc_message_new(blog->url);
	soup_xmlrpc_message_start_call(msg, "metaWeblog.newMediaObject");
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->id);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->user);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_write_string(msg, blog->pass);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_start_struct(msg);
	soup_xmlrpc_message_start_member(msg, "name");
	soup_xmlrpc_message_write_string(msg, image_name);
	soup_xmlrpc_message_end_member(msg);
	soup_xmlrpc_message_start_member(msg, "type");
	soup_xmlrpc_message_write_string(msg, mime_type);
	soup_xmlrpc_message_end_member(msg);
	soup_xmlrpc_message_start_member(msg, "bits");
	soup_xmlrpc_message_write_base64(msg, image_data, data_size);
	soup_xmlrpc_message_end_member(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	soup_message_add_header(SOUP_MESSAGE(msg)->request_headers, "Expect", "100-continue");

	soup_session_queue_message(data->session, SOUP_MESSAGE(msg), _blog_metaweblog_uploaded,
			clo);

	return TRUE;
}
